/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Verbosio XML Editor.
 *
 * The Initial Developer of the Original Code is
 * Alexander J. Vincent <ajvincent@gmail.com>.
 * Portions created by the Initial Developer are Copyright (C) 2005-2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/* namespaceRegistry.data[k] = {
 *   namespaceURI: Language namespace URI.
 *   chromeName:   Chrome package name.
 *   contentType:  Content type for the language.
 *   version:      Version of the language.
 * }
 */

const namespaceRegistry = new DataRegistry();
namespaceRegistry.requiredFields = [
  "namespaceURI",
  "chromeName",
  "contentType",
  "version"];
namespaceRegistry.addProperties({
  _isInitialized: false,

  /**
   * Look up the XML namespaces we have to work with.
   */
  init: function init() {
    if (this._isInitialized) {
      return;
    }
    var metadataDocs = verbosio.extensionDataFiles;
    for (var i = 0; i < metadataDocs.length; i++) {
      var doc = metadataDocs[i].contentDocument;
      var langElements = doc.getElementsByTagNameNS(MARKUP_NS,
                                                    "language");
      if (!langElements) {
        continue;
      }
      var lang = langElements[0];
      if (!lang) {
        continue;
      }
      if (lang.getAttribute("type") != "xml-markup") {
        continue;
      }
      var path = metadataDocs[i].location;
      var obj = {
        location:     path,
        namespaceURI: lang.getAttribute("namespaceURI"),
        chromeName:   lang.getAttribute("chrome-name"),
        contentType:  lang.getAttribute("content-type"),
        version:      lang.getAttribute("version")
      }
      this.addEntry(obj);
    }
    this._isInitialized = true;
  },

  /* This registry should not be deleted from! */
  removeEntry: function nsRegFail() {
    throw Components.interfaces.NS_ERROR_FAILURE;
  },

  removeIndex: function nsRegFail() {
    throw Components.interfaces.NS_ERROR_FAILURE;
  }
});
