/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Verbosio XML Editor.
 *
 * The Initial Developer of the Original Code is
 * Alexander J. Vincent <ajvincent@gmail.com>.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Modify the selection or the document based on information from Verbosio.
 *
 * @param aType  String defining the event type.
 * @param aEvent nsIDOMEvent dispatched for master document (may be null)
 * @param aDescIndexes1 Descendant index array based on
 */
function notifyEvent(aType, aEvent, aDescIndexes1, aDescIndexes2) {
  var wrapper = this.sourceEditor.getUserData("wrapper");
  switch (aType) {
    case "select":
      // fall through; we want to do the same as if we had a focus call.

    case "focus":
      this.sourceEditor.initialize();
      break;

    case "blur":
      var source;
      if (wrapper instanceof Components.interfaces.xeIXMLDocumentWrapper) {
        source = gDocumentPack.entityManager.getSanitizedContent(wrapper);
      } else {
        source = wrapper.wrappedSource;
      }
      if (this.sourceEditor.sourceFragment != source) {
        verbosio.advanceTxn(this.sourceEditor.transaction, wrapper);
        verbosio.advanceTxn(null, wrapper);
      }
      // do nothing, selection's already handled.
      break;

    case "mutation":
      break;

    case "repaint":
      break;

    default:
      throw new Error("notifyEvent called with invalid value!");
  }
}

function initializeView() {
  var wrapper = this.sourceEditor.getUserData("wrapper");
  assert(wrapper, "No wrapper?");
  var editor = this.sourceEditor;
  editor.sourceMarkup.value = wrapper.wrappedSource;
  editor.initialize();
}

function disableWhenInvalid() {
  var invalid = this.sourceEditor.invalidMessage;
  if (invalid) {
    documentViewer.publicURITabs.setAttribute("disabled", "true");
    documentViewer.viewTabs.setAttribute("disabled", "true");
  } else {
    documentViewer.publicURITabs.removeAttribute("disabled");
    documentViewer.viewTabs.removeAttribute("disabled");
  }
}

/**
 * Get a viewer node for a given document.
 *
 * @param aWrapper needing a viewer.
 *
 * @return clone of inspectorView element.
 *
 * @note aWrapper may be a xeIDocumentWrapper, xeITextFragmentWrapper, or a
 *       xeIXMLFragmentWrapper.
 */
function getViewerNode(aWrapper) {
  dump("Getting source edit viewer for " + aWrapper + "\n");
  var hboxView = document.createElementNS(XUL_NS, "xul:hbox");
  var viewer = document.createElementNS(XUL_NS, "xul:source-editor");
  hboxView.appendChild(viewer);

  hboxView.setAttributeNS(null, "scopeid", uuid);
  hboxView.sourceEditor = viewer;
  viewer.setUserData("wrapper", aWrapper, null);
  hboxView.notifyEvent = notifyEvent;
  hboxView.initializeView = initializeView;
  hboxView.updateAcceptDisabled = disableWhenInvalid;

  if (isInstance(aWrapper, "xeIXMLDocumentWrapper")) {
    viewer.setAttribute("type", "xeIXMLDocumentWrapper");
    return hboxView;
  }

  if (isInstance(aWrapper, "xeITextDocumentWrapper")) {
    viewer.setAttribute("type", "xeITextDocumentWrapper");
    return hboxView;
  }

  if (isInstance(aWrapper, "xeIXMLFragmentWrapper")) {
    viewer.setAttribute("type", "xeIXMLFragmentWrapper");
    return hboxView;
  }

  if (isInstance(aWrapper, "xeITextFragmentWrapper")) {
    viewer.setAttribute("type", "xeITextFragmentWrapper");
    return hboxView;
  }

  return null;
}

// XXX ajvincent Localize!
window.addEventListener("load", function addInspectorTab(evt) {
  documentViewer.appendViewerTab("Source code", uuid, "s");
}, true);
